/*
   pravidla:
     1s=1000ms
     Uint32 SDL_GetTicks() vracia pocet ms od startu SDL,
     po zhruba 49 dnoch dojde k prekroceniu 32bit limitu
     a citac sa vynuluje
   uzitocne vzorce:
     rychlost  v=s/t [m/s]  draha  s=vt [m]
               v=at [m/s]
     zrychlenie             a=s/t2 [m/s2]
     gravitacne zrychlenie  g=v/t [m/s2], g na Zemi=9.81 m/s2
                            g=s/t2  draha  s=gt2 [m]
*/
// hlavickove subory *******************************************************
#include <math.h>
// GL a GLU kniznica
#include <GL/gl.h>
#include <GL/glu.h>
// funkcie SDL kniznice
#include <SDL/SDL.h>

// vlastny .h subor
#include "mAvatar.h"

// konstruktor *************************************************************
mAvatar::mAvatar()
{
 scr_w=0.0f; scr_h=0.0f; // pociatocne rozmery okna

 cam_reset(); // reset parametrov kamery
}

// destruktor **************************************************************
mAvatar::~mAvatar()
{

}

// reset parametrov kamery *************************************************
void mAvatar::cam_reset(void)
{
 // vyska oci AVATARa = y-ova suradnica kamery
 AVATAR_eye=1.7f;
 // sirka AVATARa
 AVATAR_width=1.0f;

 // naplnenie matice kamery pociatocnymi hodnotami
 for (t_a=0;t_a<16;t_a++) matrix[t_a]=0.0f;
 matrix[0]=1.0f; matrix[5]=1.0f; matrix[10]=1.0f; matrix[15]=1.0f;
 // nastavenie parametrov kamery
 posx=0.0f; posy=-AVATAR_eye; posz=0.0f;
 oldx=posx; oldy=posy; oldz=posz;
 targetx=0.0f; targety=0.0f; targetz=1.0f;
 movx=0.0f; movy=0.0f; movz=0.0f;
 angx=0.0f; angy=0.0f; angz=0.0f;
 rotx=0.0f; roty=0.0f; rotz=0.0f;
 speed=4.0f; slide=4.0f; altit=4.0f;
 acceleration=2.0f; gravity=9.81f;
 viewnear=0.1f; viewfar=1000.0f;
 FOV=60.0f;
 JumpPower=0.0f;

 // pociatocne zresetovanie posunu
 goForward=0.0f; goBack=0.0f; goUp=0.0f; goDown=0.0f; goRight=0.0f; goLeft=0.0f;
 goJump=0.0f; goFast=0.0f;

 // strelba
 isFire1=0;
 isFire2=0;

 // sposob vypoctu pohybu,0=chodza,1=let
 movetype=1;

 // reset casu
 cam_resettime();
}

// reset casu **************************************************************
void mAvatar::cam_resettime(void)
{
 // reset casu
 actTime=SDL_GetTicks();
 oldTime=actTime;
 difTime=(GLfloat) (actTime-oldTime);

 gravityTime=0.0f;
 cam_fps=0.0f;
}

// reinicializacia kamery **************************************************
void mAvatar::cam_reinit(GLfloat W,GLfloat H)
{
 glMatrixMode(GL_PROJECTION); // vyber projekcnej matice
 glLoadIdentity();            // zresetovanie projekcnej matice

 if (H==0.0f) H=1.0f;
 gluPerspective(FOV,W/H,viewnear,viewfar); // nastavenie perspektivneho premietania
 // FOV - zorny uhol pohladu, W/H - aspect - pomer vysky a sirky zobrazovanej plochy
 // viewnear - minimalna viditelna vzdialenost, viewfar - maximalna viditelna vzdialenost
 glViewport(0,0,(GLuint) W,(GLuint) H);    // nastavenie viditelnej plochy okna

 // TRANSFORMUJEME SCENU NIE KAMERU!
 glMatrixMode(GL_MODELVIEW); // vyber modelovej matice
 glLoadIdentity();           // zresetovanie modelovej matice

 scr_w=W; scr_h=H; // ulozenie rozmerov okna
}

// nastavenie perspektivneho premietania ***********************************
void mAvatar::cam_set_perspective(void)
{
 gluPerspective(FOV,scr_w/scr_h,viewnear,viewfar); // nastavenie perspektivneho premietania
}

// rotacia kamery **********************************************************
void mAvatar::cam_rotate_xy(GLfloat X,GLfloat Y)
{
 // nastavenie rotacie, hodnoty su korigovane pre dosiahnutie realneho vnemu
 rotx=X/10.0f; roty=Y/10.0f;
}

// rotacia kamery **********************************************************
void mAvatar::cam_rotate_zl(GLfloat Z)
{
 // nastavenie rotacie, hodnota je korigovana pre dosiahnutie realneho vnemu
 rotz=-Z;
}

// rotacia kamery **********************************************************
void mAvatar::cam_rotate_zr(GLfloat Z)
{
 // nastavenie rotacie, hodnota je korigovana pre dosiahnutie realneho vnemu
 rotz+=Z;
}

// posun kamery ************************************************************
void mAvatar::cam_move(void)
{
 // cas... pohyb je zavisly od casu, rotacia nie...
 actTime=SDL_GetTicks();
 difTime=((GLfloat) (actTime-oldTime))/1000.0f;
 oldTime=actTime;

 // INFO: vypis casoveho okamihu uplynuteho medzi vykreslenim dvoch scen
 //fprintf(stderr,"difTime=%f\n",difTime);

 // pocitadlo FPS
 if (difTime==0.0f) difTime=1.0f;
 cam_fps=1.0f/difTime;

 // pohyb kamery, ak su stlacene sucasne dve klavesy, AVATAR sa nehybe
 goFast=50.0f*goFast+1.0f;
 movz=(goFast)*(goForward-goBack)*speed*difTime; // dopredu a dozadu
 movx=(goFast)*(goRight-goLeft)*slide*difTime;   // vlavo a vpravo
 movy=(goFast)*(goUp-goDown)*altit*difTime;      // hore a dole

 // pri type pohybu chodza posobi gravitacia
 if (movetype==0) {
  if (gravityTime==0.0f) { // ak je AVATAR na zemi
   if (cam_fps==0.0f) cam_fps=1.0f;       // kontrola na FPS=0!
   JumpPower=(goJump)*sqrt(1.0f/cam_fps); // vypocet sily skoku (zavisle od FPS!)
  }

  // zvacsovanie sa casu trvania skoku/posobenia gravitacie
  gravityTime+=difTime;

  // vypocet klesania v zavislosti od skoku a gravitacie
  movy=JumpPower-gravity*gravityTime/20.0f;
 }

 // negacia posunu
 movx*=-1.0f;  movy*=-1.0f;  movz*=-1.0f;

 // rozdiel poloh AVATARa nemoze byt vacsi ako jeho sirka
 // nakolko nastanu chyby pri vypocte kolizie
 if ((movx>0.0f)&&(movx>AVATAR_width)) movx=AVATAR_width;
 else if (movx<-AVATAR_width) movx=-AVATAR_width;
 if ((movy>0.0f)&&(movy>AVATAR_width)) movy=AVATAR_width;
 else if (movy<-AVATAR_width) movy=-AVATAR_width;
 if ((movz>0.0f)&&(movz>AVATAR_width)) movz=AVATAR_width;
 else if (movz<-AVATAR_width) movz=-AVATAR_width;

 // zrusenie dalsieho mozneho posunu alebo skoku
 goForward=0.0f; goBack=0.0f; goUp=0.0f; goDown=0.0f; goRight=0.0f; goLeft=0.0f;
 goJump=0.0f; goFast=0.0f;
}

// aktualizacia nastaveni kamery *******************************************
void mAvatar::cam_update(void)
{
 // ulozenie povodnej polohy
 oldx=posx; oldy=posy; oldz=posz;

 // SPECIALNY PRISTUP K ROTACII A NASLEDNEMU POSUNU KAMERY!
 glMatrixMode(GL_MODELVIEW);
 glLoadMatrixf(matrix); // modelova matica sa nahradi maticou kamery

 switch (movetype) {
  case 0:  // chodza
   // ROTACIA
   glRotatef(rotx,matrix[0],matrix[4],matrix[8]);
   glRotatef(roty,0.0f,1.0f,0.0f);
   // nastavenie noveho otocenia
   angx+=rotx; angy+=roty; angz+=rotz;
   // ulozenie udajov do matice kamery
   glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
   // nastavenie novej polohy - POSUN sa vykona neskor v ogl_Draw
   posx+=matrix[0]*movx-matrix[2]*movz;
   posy+=movy;
   posz+=matrix[8]*movx-matrix[10]*movz;
   break;
  case 1:  // let
   // ROTACIA
   glRotatef(rotx,matrix[0],matrix[4],matrix[8]);
   glRotatef(roty,matrix[1],matrix[5],matrix[9]);
   glRotatef(rotz,matrix[2],matrix[6],matrix[10]);
   // nastavenie noveho otocenia
   angx+=rotx; angy+=roty; angz+=rotz;
   // ulozenie udajov do matice kamery
   glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
   // nastavenie novej polohy - POSUN sa vykona neskor v ogl_Draw
   posx+=matrix[0]*movx+matrix[1]*movy-matrix[2]*movz;
   posy+=matrix[4]*movx+matrix[5]*movy-matrix[6]*movz;
   posz+=matrix[8]*movx+matrix[9]*movy-matrix[10]*movz;
   break;
  default: // dalsie typy pohybu...
   break;
 }

 // zrusenie dalsej moznej rotacie
 rotx=0.0f; roty=0.0f; rotz=0.0f;
 // nastavenie smeroveho vektora kamery
 targetx=matrix[2]; targety=matrix[6]; targetz=matrix[10];
}

// ziskanie casoveho prirastku *********************************************
GLfloat mAvatar::cam_gettime(void)
{
 return(difTime);
}
