// hlavickove subory *******************************************************
// systemove kniznice
#include <math.h>
// GL,GLU a GLEXT kniznice
#include <GL/gl.h>
// funkcie SDL kniznice (pouzije sa subor /usr/lib/libSDL.so)
#include <SDL/SDL.h>

// vlastny .h subor
#include "mMD3model.h"

// tabulka predpripravenych normal
float nxyz[256][256][3];

// test dostupnosti suboru *************************************************
int CheckFile(char *filename)
{
 FILE *f;
 f=fopen(filename,"rb");

 if (!f) return(1); // subor nemozno otvorit
 else { fclose(f); return(0); }
}

// pre nacitanie retazca zo suboru
char str_token[1024];

// nacitanie retazca zo suboru *********************************************
char *ParseString(char *data)
{
 int c,len;

 len=0;
 str_token[0]=0;

 if (!data) return(NULL);

skipwhite:
 // preskocenie prazdnych miest
 while ((c=*data)<=' ') {
  if (c==0) return(NULL); // EOF
  data++;
 }
 // preskocenie komentarov
 if (c=='/' && data[1]=='/') {
  while (*data && *data!='\n') data++;
  goto skipwhite;
 }

 // retazec uzavrety v uvodzovkach
 if (c=='\"') {
  data++;
  do {
   c=*data++;
   if (c=='\"') {
    str_token[len]=0;
    return(data);
   }
   str_token[len]=c; len++;
  } while (1);
 }

 // jednotlive znaky
 if (c=='{' || c=='}' || c==')' || c=='(' || c=='\'' || c==':' || c==',' || c=='/') {
  str_token[len]=c; len++;
  str_token[len]=0;
  return(data+1);
 }

 // regularne slovo
 do {
  str_token[len]=c; len++;
  data++;
  c=*data;
  if (c=='{' || c=='}'|| c==')'|| c=='(' || c=='\'' || c==':' || c==',' || c=='/') break;
 } while (c>32);

 str_token[len]=0;

 return(data);
}

// prechod cislami ulozenymi v retazci *************************************
int ParsNum(char *data)
{
 while (*data) {
  if (*data<'0' || *data>'9') return(0);
  data++;
 }

 return(1);
}

// MD3Model ****************************************************************

// konstruktor MD3 model ***************************************************
MD3Model::MD3Model(void)
{
 frames=NULL;   // snimky
 tags=NULL;     // znacky
 links=NULL;    // prepojenia
 surfaces=NULL; // sietove modely

 frame=0;
 nextframe=0;
 ipol=0.0f;
}

// destruktor MD3 model ****************************************************
MD3Model::~MD3Model(void)
{
 // zrusenie snimok
 if (frames) free(frames);
 // zrusenie znaciek
 if (tags) free(tags);
 // zrusenie prepojeni
 if (links) {
  for (i=0;i<header.numTags;i++) { if (links[i]) links[i]=NULL; }
  free(links);
 }
 // zrusenie sietovych modelov
 if (surfaces) {
  for (i=0;i<header.numSurfaces;i++) {
   if (surfaces[i].shaders) free(surfaces[i].shaders);
   if (surfaces[i].triangles) free(surfaces[i].triangles);
   if (surfaces[i].texcoords) free(surfaces[i].texcoords);
   if (surfaces[i].vertices) free(surfaces[i].vertices);
  }
  free(surfaces);
 }

 frames=NULL;
 tags=NULL;
 links=NULL;
 surfaces=NULL;
}

// nacitanie obsahu MD3 suboru *********************************************
int MD3Model::LoadModel(char *filename)
{
 if (CheckFile(filename)) {
  fprintf(stderr,"Nemozno najst subor %s!\n",filename); return(1);
 }

 // INFO: cesta a nazov MD3 suboru
 fprintf(stderr,"Najdeny subor: %s ",filename);

 // otvorenie MD3 suboru
 md3file=fopen(filename,"rb");

 // nacitanie hlavicky MD3 suboru
 fread(&header,1,sizeof(md3_header_t),md3file);
 // INFO: hlavicka MD3 suboru
 fprintf(stderr,"ident: %c%c%c%c\n",header.ident[0],header.ident[1],header.ident[2],header.ident[3]);
 fprintf(stderr,"frames: %u\n",header.numFrames);
 fprintf(stderr,"tags: %u\n",header.numTags);
 fprintf(stderr,"surfaces: %u\n",header.numSurfaces);
 fprintf(stderr,"skins: %u\n",header.numSkins);
 fprintf(stderr,"name: %s\n",header.name);

 // test verzie MD3 suboru
 if (header.version!=15) {
  fclose(md3file);
  fprintf(stderr,"Nepodporovana verzia %u!",header.version); return(1);
 }

 // nacitanie snimkov
 frames=(frame_t *) malloc(sizeof(frame_t)*header.numFrames);
 fread(frames,sizeof(frame_t),header.numFrames,md3file);

 // nacitanie znaciek
 tags=(tag_t *) malloc(sizeof(tag_t)*header.numFrames*header.numTags);
 fread(tags,sizeof(tag_t),header.numFrames*header.numTags,md3file);

 // INFO: znacky MD3 suboru
 for (i=0;i<header.numTags;i++) fprintf(stderr,"znacka: %u %s\n",i,tags[i].name);

 // nastavenie prepojeni
 links=(MD3Model **) malloc(sizeof(MD3Model)*header.numTags);
 for (i=0;i<header.numTags;i++) links[i]=NULL;

 // nacitanie sietovych modelov
 surfaces=(surface_t *) malloc(sizeof(surface_t)*header.numSurfaces);

 // ziskanie aktualnej pozicie v subore
 long surface_offset=ftell(md3file);

 // nacitanie sietovych modelov
 for (i=0;i<header.numSurfaces;i++) {
  // presun na patricnu poziciu v subore
  fseek(md3file,surface_offset,SEEK_SET);
  // hlavicka sietoveho modelu
  fread(&surfaces[i].surface_h,sizeof(surface_header_t),1,md3file);
  // povrchy
  surfaces[i].shaders=(shader_t *) malloc(sizeof(shader_t)*surfaces[i].surface_h.numShaders);
  fread(surfaces[i].shaders,sizeof(shader_t),surfaces[i].surface_h.numShaders,md3file);
  // indexy na vrcholy trojuholnikov
  fseek(md3file,surface_offset+surfaces[i].surface_h.triangles_start,SEEK_SET);
  surfaces[i].triangles=(triangle_t *) malloc(sizeof(triangle_t)*surfaces[i].surface_h.numTriangles);
  fread(surfaces[i].triangles,sizeof(triangle_t),surfaces[i].surface_h.numTriangles,md3file);
  // texturovacie suradnice
  fseek(md3file,surface_offset+surfaces[i].surface_h.st_start,SEEK_SET);
  surfaces[i].texcoords=(texcoord_t *) malloc(sizeof(texcoord_t)*surfaces[i].surface_h.numVerts);
  fread(surfaces[i].texcoords,sizeof(texcoord_t),surfaces[i].surface_h.numVerts,md3file);
  // vrcholy
  fseek(md3file,surface_offset+surfaces[i].surface_h.xyznormal_start,SEEK_SET);
  surfaces[i].vertices=(vertex_t *) malloc(sizeof(vertex_t)*surfaces[i].surface_h.numVerts*surfaces[i].surface_h.numFrames);
  fread(surfaces[i].vertices,sizeof(vertex_t),surfaces[i].surface_h.numFrames*surfaces[i].surface_h.numVerts,md3file);
  // presun na dalsi sietovy model
  surface_offset+=surfaces[i].surface_h.size;
 }

 // zatvorenie MD3 suboru
 fclose(md3file);

 // posledny snimok pre Q3 model je header.numFrames-1
 header.numFrames-=1;

 return(0);
}

// nacitanie obrazkov pre povrchy sietovych modelov ************************
int MD3Model::LoadSkin(char *filename,char *imagepath)
{
 FILE *f;
 char *text_line,*buffer;
 char name_surface[1024],name_image[1024],path[1024];
 unsigned char issurface,istag,isimage;

 if (CheckFile(filename)) {
  fprintf(stderr,"Nemozno najst subor %s!\n",filename); return(1);
 }
 // otvorenie suboru
 f=fopen(filename,"rt");
 buffer=(char *) malloc(1024);
 // nacitanie obrazkov
 while (!feof(f)) {
  istag=0;
  issurface=0; isimage=0;
  text_line=buffer;
  text_line[0]='\0';
  fscanf(f,"%s",text_line);

  if (text_line && text_line[0]) {
   // nacitanie znacky,nazvu povrchu,nazvu obrazka
   while (text_line) {
    text_line=ParseString(text_line);

    if (strstr(str_token,"tag_") && issurface==0 && istag==0) { istag=1; }
    else if (issurface==0 && istag==0) { strcpy(name_surface,str_token); issurface=1; }
         else if ((strstr(str_token,".tga")||strstr(str_token,".jpg")) && issurface==1 && istag==0)
              { strcpy(name_image,str_token); isimage=1; }
   }

   // nacitanie obrazka
   if (issurface && isimage) {
    fprintf(stderr,"%s...%s\n",name_surface,name_image);

    if (strstr(name_image,".tga") || strstr(name_image,".jpg")) {
     sprintf(path,"data/model/%s/%s",imagepath,name_image);
     // tesst na priesvitnost textury
     if (strstr(name_image,"TRT")) i=tt.LoadTexture(path,T_MIPMAP|T_TRANSPARENT);
     else i=tt.LoadTexture(path,T_MIPMAP);
     // nastavenie textury pre povrch sietoveho modelu
     if (!i) SetSurfaceTexture(name_surface,tt.texture_id[tt.texture_count-1]);
    }
   }
  }
 }

 free(buffer);
 // zatvorenie suboru
 fclose(f);

 return(0);
}

// nastavenie textury pre povrchy sietovych modelov ************************
int MD3Model::SetSurfaceTexture(char *surfacename,unsigned int TID)
{
 for (i=0;i<header.numSurfaces;i++) {
  if (!strcmp(surfaces[i].surface_h.name,surfacename)) {
   //fprintf(stderr,"surface: %s texture id: %u\n",surfaces[i].surface_h.name,TID);
   surfaces[i].tex_id=TID;
   surfaces[i].is_texture=1;
   return(i);
  }
 }
 return(0);
}
/*
// nastavenie textury pre vsetky povrchy sietovych modelov *****************
void MD3Model::SetAllSurfacesTexure(unsigned int TID)
{
 for (i=0;i<header.numSurfaces;i++) {
  surfaces[i].tex_id=TID;
  surfaces[i].is_texture=1;
 }
}
*/

// prepojenia modelov ******************************************************
int MD3Model::LinkModel(char *tagname,MD3Model *mod)
{
 // v pripade,ze sa nazov znacky rovna zadanemu nazvu,
 // k prepojeniam aktualneho modelu sa doplni odkaz na novy model

 // pre vsetky znacky aktualneho modelu
 for (i=0;i<header.numTags;i++) {
  // ak sa nazov znacky rovna zadanemu nazvu
  if (!strcmp(tags[i].name,tagname)) {
   // vznikne nove prepojenie
   links[i]=mod;
   return(i);
  }
 }

 return(0);
}
/*
// zrusenie prepojeni modelov **********************************************
int MD3Model::UnLinkModel(char *tagname)
{
 // v pripade,ze sa nazov znacky rovna zadanemu nazvu,
 // z prepojeni aktualneho modelu sa odstrani odkaz na iny model

 // pre vsetky znacky aktualneho modelu
 for (i=0;i<header.numTags;i++) {
  // ak sa nazov znacky rovna zadanemu nazvu
  if (!strcmp(tags[i].name,tagname)) {
   // zrusi sa prepojenie
   links[i]=NULL;
   return(0);
  }
 }

 return(0);
}
*/
// nastavenie informacii o snimku ******************************************
void MD3Model::SetFrameInfo(int f,int n,float p)
{
 frame=f;
 nextframe=n;

 if (p<0.0f || p>1.0f) p=0.0f;
 ipol=p;
}

// ziskanie informacii o snimku ********************************************
void MD3Model::GetFrameInfo(int *f,int *n,float *p)
{
 *f=frame;
 *n=nextframe;
 *p=ipol;
}

// vykreslenie 1 snimku ****************************************************
void MD3Model::DrawModelOneFrame(int curFrame)
{
 // pre vsetky povrchy
 for (cur_surface=0;cur_surface<header.numSurfaces;cur_surface++) {
  // posun na prvy vrchol
  cur_offset_vertex=curFrame*surfaces[cur_surface].surface_h.numVerts;
  // pocet trojuholnikov
  tri_num=surfaces[cur_surface].surface_h.numTriangles;
  // povrch ma texturu
  if (surfaces[cur_surface].is_texture) glBindTexture(GL_TEXTURE_2D,surfaces[cur_surface].tex_id);
  // pre vsetky trojuholniky
  for (i=0;i<tri_num;i++) {
   glBegin(GL_TRIANGLES);
   // pre vsetky vrcholy trojuholnikov
   for (j=0;j<3;j++) {
    // index vrcholu
    cur_vertex=surfaces[cur_surface].triangles[i].indexes[j];
    // texturovacie suradnice
    s=surfaces[cur_surface].texcoords[cur_vertex].st[0];
    t=surfaces[cur_surface].texcoords[cur_vertex].st[1];
    // normaly
    au=(int) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].n[0];
    av=(int) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].n[1];
    // dekodovanie normal
    n[0]=nxyz[au][av][0];
    n[1]=nxyz[au][av][1];
    n[2]=nxyz[au][av][2];
    //  vrcholy
    v[0]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[0]/64.0f;
    v[1]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[1]/64.0f;
    v[2]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[2]/64.0f;

    glTexCoord2f(s,t);
    glNormal3f(n[0],n[1],n[2]);
    glVertex3f(v[0],v[1],v[2]);
   }

   glEnd();
  }
 }
}

// vykreslenie interpolovaneho snimku **************************************
void MD3Model::DrawModelIFrame(int curFrame,int nextFrame,float p)
{
 // pre vsetky povrchy
 for (cur_surface=0;cur_surface<header.numSurfaces;cur_surface++) {
  // posun na prvy vrchol
  cur_offset_vertex=curFrame*surfaces[cur_surface].surface_h.numVerts;
  // interpolacia
  next_cur_offset_vertex=nextFrame*surfaces[cur_surface].surface_h.numVerts;
  // pocet trojuholnikov
  tri_num=surfaces[cur_surface].surface_h.numTriangles;
  // povrch ma texturu
  if (surfaces[cur_surface].is_texture) glBindTexture(GL_TEXTURE_2D,surfaces[cur_surface].tex_id);
  //fprintf(stderr,"surface: %s texture id: %u\n",surfaces[cur_surface].surface_h.name,surfaces[cur_surface].tex_id);

  // pre vsetky trojuholniky
  for (i=0;i<tri_num;i++) {
   glBegin(GL_TRIANGLES);
   // pre vsetky vrcholy trojuholnikov
   for (j=0;j<3;j++) {
    // index vrcholu
    cur_vertex=surfaces[cur_surface].triangles[i].indexes[j];
    // texturovacie suradnice
    s=surfaces[cur_surface].texcoords[cur_vertex].st[0];
    t=surfaces[cur_surface].texcoords[cur_vertex].st[1];
    // normaly
    au=(int) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].n[0];
    av=(int) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].n[1];
    n[0]=nxyz[au][av][0];
    n[1]=nxyz[au][av][1];
    n[2]=nxyz[au][av][2];
     // interpolacia
     au=(int) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].n[0];
     av=(int) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].n[1];
     next_n[0]=nxyz[au][av][0];
     next_n[1]=nxyz[au][av][1];
     next_n[2]=nxyz[au][av][2];
    //  vrcholy
    v[0]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[0]/64.0f;
    v[1]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[1]/64.0f;
    v[2]=(float) surfaces[cur_surface].vertices[cur_offset_vertex+cur_vertex].v[2]/64.0f;
     // interpolacia
     next_v[0]=(float) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].v[0]/64.0f;
     next_v[1]=(float) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].v[1]/64.0f;
     next_v[2]=(float) surfaces[cur_surface].vertices[next_cur_offset_vertex+cur_vertex].v[2]/64.0f;

    glTexCoord2f(s,t);
    glNormal3f(n[0]+p*(next_n[0]-n[0]),
               n[1]+p*(next_n[1]-n[1]),
               n[2]+p*(next_n[2]-n[2]));
    glVertex3f(v[0]+p*(next_v[0]-v[0]),
               v[1]+p*(next_v[1]-v[1]),
               v[2]+p*(next_v[2]-v[2]));
   }

   glEnd();
  }
 }
}

// vykreslenie modelu (1 snimok alebo interpolovany snimok) ****************
void MD3Model::DrawModel(void)
{
 //DrawModelOneFrame(frame);
 DrawModelIFrame(frame,nextframe,ipol);
}

// MD3Player ***************************************************************

// konstruktor
MD3Player::MD3Player(void)
{
 // plynutie casu
 aTime=0.0f;
 l_curFrame=0; u_curFrame=0;
 l_nextFrame=0; u_nextFrame=0;
 l_curTime=0.0f; u_curTime=0.0f;
 l_oldTime=0.0f; u_oldTime=0.0f;

 for (i=0;i<MAX_ANIMATIONS;i++) {
  anim[i].first_frame=0;
  anim[i].num_frames=0;
  anim[i].looping_frames=0;
  anim[i].frames_per_second=0;
 }
}

// destruktor
MD3Player::~MD3Player(void)
{
}

// inicializacia MD3 postavy ***********************************************
void MD3Player::Init()
{
 // dekodovanie normal
 double lat,lng;
 // priprava tabulky s predpripravenymi normalami
 for (i=0;i<256;i++) {
  for (j=0;j<256;j++) {
   lat=i*2.0f*M_PI/255.0f;
   lng=j*2.0f*M_PI/255.0f;

   nxyz[i][j][0]=(float) (cos(lat)*sin(lng));
   nxyz[i][j][1]=(float) (sin(lat)*sin(lng));
   nxyz[i][j][2]=(float) cos(lng);
  }
 }

 // nacitanie MD3 postavy + textur + zbrane
 LoadPlayer("sarge","default","railgun");
}

// nacitanie MD3 postavy ***************************************************
void MD3Player::LoadPlayer(char *dirname,char *skin,char *weapon)
{
 char l_mpath[1024],u_mpath[1024],h_mpath[1024];
 char l_spath[1024],u_spath[1024],h_spath[1024];
 char a_path[1024];
 // zbran
 char w_mpath[1024],w_spath[1024];

 // casti MD3 postavy
 sprintf(l_mpath,"data/model/%s/lower.md3",dirname);
 sprintf(u_mpath,"data/model/%s/upper.md3",dirname);
 sprintf(h_mpath,"data/model/%s/head.md3",dirname);
 // zbran
 if (weapon!=NULL) sprintf(w_mpath,"data/model/%s/%s.md3",weapon,weapon);
 // casti MD3 povrchu
 sprintf(l_spath,"data/model/%s/lower_%s.skin",dirname,skin);
 sprintf(u_spath,"data/model/%s/upper_%s.skin",dirname,skin);
 sprintf(h_spath,"data/model/%s/head_%s.skin",dirname,skin);
 // zbran
 if (weapon!=NULL) sprintf(w_spath,"data/model/%s/%s.skin",weapon,weapon);
 // animacia
 sprintf(a_path,"data/model/%s/animation.cfg",dirname);
 // nacitanie modelov
 l.LoadModel(l_mpath); // lower
 u.LoadModel(u_mpath); // upper
 h.LoadModel(h_mpath); // head
 if (weapon!=NULL) w.LoadModel(w_mpath); // weapon
 // nacitanie povrchov
 l.LoadSkin(l_spath,dirname); // lower
 u.LoadSkin(u_spath,dirname); // upper
 h.LoadSkin(h_spath,dirname); // head
 if (weapon!=NULL) w.LoadSkin(w_spath,weapon);  // weapon
 // prepojenie modelov
 l.LinkModel("tag_torso",&u);  // lower -> upper (naviazany na tag_torso)
 u.LinkModel("tag_head",&h);   // upper -> head (naviazany na tag_head)
 if (weapon!=NULL) u.LinkModel("tag_weapon",&w); // upper -> weapon (naviazany na tag_weapon)
/*
 // INFO: prepojenia medzi modelmi
 fprintf(stderr,"lower:\n");
 for (i=0;i<l.header.numTags;i++) {
  if (l.links[i]) fprintf(stderr,"link %u: P",i); else fprintf(stderr,"link %u: NULL",i);
  fprintf(stderr,"\n");
 }
 fprintf(stderr,"upper:\n");
 for (i=0;i<u.header.numTags;i++) {
  if (u.links[i]) fprintf(stderr,"link %u: P",i); else fprintf(stderr,"link %u: NULL",i);
  fprintf(stderr,"\n");
 }
 fprintf(stderr,"head:\n");
 for (i=0;i<h.header.numTags;i++) {
  if (h.links[i]) fprintf(stderr,"link %u: P",i); else fprintf(stderr,"link %u: NULL",i);
  fprintf(stderr,"\n");
 }
*/
 // nacitanie animacie
 LoadAnim(a_path);
 // nastavenie parametrov animacie
 SetAnim(TORSO_STAND);
}

// nacitanie animacie ******************************************************
int MD3Player::LoadAnim(char *filename)
{
 FILE *f;
 char *text_line,*buffer;
 int c;
 int a=0,t=0;
 int issex,isheadoffset,isfootsteps,skip;

 if (CheckFile(filename)) {
  fprintf(stderr,"Nemozno najst subor %s!\n",filename); return(1);
 }
 // otvorenie suboru
 f=fopen(filename,"rt");
 buffer=(char *) malloc(1024);
 // nacitanie obsahu suboru
 while (!feof(f)) {
  c=fgetc(f);

  i=0;
  text_line=buffer;
  text_line[0]=0;
  // preskocenie 10=new line,32=space,-1=EOF
  while (c!=10 && c!=32 && c!=-1) {
   text_line[i]=c; i++;
   c=fgetc(f);
  }
  text_line[i]='\0';

  t=0; // 4 ciselne udaje patriace animacii
  issex=0;
  isheadoffset=0;
  isfootsteps=0;
  while (text_line) {
   text_line=ParseString(text_line);
   if (str_token && str_token[0]!=0) {
    if (strstr(str_token,"sex")) issex=1;
    if (strstr(str_token,"headoffset")) isheadoffset=1;
    if (strstr(str_token,"footsteps")) isfootsteps=1;

    // ak je najdene cislo, naplni sa jednen zo 4 ciselnych udajov
    // patriacich animacii
    if (ParsNum(str_token) && !issex && !isheadoffset && !isfootsteps) {
     if (t==0) { anim[a].first_frame=atoi(str_token); t++; }
     else if (t==1) { anim[a].num_frames=atoi(str_token); t++; }
          else if (t==2) { anim[a].looping_frames=atoi(str_token); t++; }
               else if (t==3) { anim[a].frames_per_second=atoi(str_token); t++; }

     if (t==4) a++;
    }
   }
  }
 }

 // v subore upper.md3 su definovane pohyby 0-4 + nasleduju pohyby pre TORSO
 // v subore lower.md3 su definovane pohyby 0-4 + nasleduju pohyby pre LEGS
 // preto je potrebne pre pobyhy noh posunut prvy snimok na rovnaku poziciu
 // aku ma prvy snimok pohybu TORSO_GESTURE
 skip=anim[LEGS_WALKCR].first_frame-anim[TORSO_GESTURE].first_frame;
 for (a=LEGS_WALKCR;a<MAX_ANIMATIONS;a++) anim[a].first_frame-=skip;

 // uprava poctu snimkov
 for (a=0;a<MAX_ANIMATIONS;a++) {
  if (anim[a].num_frames>0) anim[a].num_frames-=1;
  // INFO: prvy snimok daneho druhu pohybu
  //fprintf(stderr,"anim[%u]=%u\n",a,anim[a].first_frame);
 }

 free(buffer);
 // zatvorenie suboru
 fclose(f);

 return(0);
}

// nastavenie parametrov animacie ******************************************
void MD3Player::SetAnim(int a)
{
 if (a>=0 && a<=5) { // upper & lower
  anim_l=a;
  anim_u=a;
 }
 else if (a>5 && a<13) { // upper, lower -> idle
       anim_l=LEGS_IDLE;
       anim_u=a;
      }
      else if (a>=13 && a<=24) { // lower, upper -> last anim
            anim_l=a;
            anim_u=anim_u;
           }

 // first frame
 l_startFrame=anim[anim_l].first_frame;
 u_startFrame=anim[anim_u].first_frame;
 // nextframe
 l_nextFrame=anim[anim_l].first_frame;
 u_nextFrame=anim[anim_u].first_frame;
 // num frames
 l_endFrame=anim[anim_l].first_frame+anim[anim_l].num_frames;
 u_endFrame=anim[anim_u].first_frame+anim[anim_u].num_frames;
 // frames per second
 l_fps=(float) anim[anim_l].frames_per_second;
 u_fps=(float) anim[anim_u].frames_per_second;
}

// vykreslenie MD3 postavy *************************************************
void MD3Player::Render(void)
{
 // plynutie casu
 aTime=SDL_GetTicks()/1000.0f;
 // vypocet parametrov snimku pre cast lower
 l_curTime=aTime;
 if (l_curTime-l_oldTime>1.0f/l_fps) {
  l_curFrame=l_nextFrame;
  l_nextFrame++;
  if (l_nextFrame>l_endFrame) l_nextFrame=l_startFrame;
  l_oldTime=l_curTime;
 }
 // vypocet parametrov snimku pre cast upper
 u_curTime=aTime;
 if (u_curTime-u_oldTime>1.0f/u_fps) {
  u_curFrame=u_nextFrame;
  u_nextFrame++;
  if (u_nextFrame>u_endFrame) u_nextFrame=u_startFrame;
  u_oldTime=u_curTime;
 }
 // nastavenie parametrov snimku pre cast lower
 l.SetFrameInfo(l_curFrame,l_nextFrame,l_fps*(l_curTime-l_oldTime));
 //fprintf(stderr,"%u %u\n",l_curFrame,l_nextFrame);
 // nastavenie parametrov snimku pre cast upper
 u.SetFrameInfo(u_curFrame,u_nextFrame,u_fps*(u_curTime-u_oldTime));
 //fprintf(stderr,"%u %u\n",u_curFrame,u_nextFrame);

 glPushMatrix();
  glTranslatef(0.0f,0.0f,-100.0f);
  glRotatef(-90.0f,1.0f,0.0f,0.0f);
  glRotatef(-100.0f,0.0f,0.0f,1.0f);
  glPushMatrix();

   // vykreslenie vsetkych casti postavy - rekurzivna funkcia !!!
   DrawPlayer(&l); // zacne sa castou lower

  glPopMatrix();
 glPopMatrix();
}

// vykreslenie vsetkych casti postavy - rekurzivna funkcia !!! *************
void MD3Player::DrawPlayer(MD3Model *mod)
{
 int i;

 // vykreslenie modelu
 mod->DrawModel();

 // vsetky casti postavy, resp. vsetky modely priradene
 // znackam aktualneho modelu
 for (i=0;i<mod->header.numTags;i++) {
  link=mod->links[i];

  if (link) {
   // interpolacia
   p=mod->ipol;

   // poloha
   position=mod->tags[mod->frame*mod->header.numTags+i].origin;
   // natocenie
   rotation=&mod->tags[mod->frame*mod->header.numTags+i].axis[0][0];

   // buduca poloha
   next_position=mod->tags[mod->nextframe*mod->header.numTags+i].origin;
   // buduce natocenie
   next_rotation=&mod->tags[mod->nextframe*mod->header.numTags+i].axis[0][0];

   // naplnenie matice (rotacia + posun)
   m[0]=rotation[0]; m[4]=rotation[3]; m[8] =rotation[6]; m[12]=position[0]+p*(next_position[0]-position[0]);
   m[1]=rotation[1]; m[5]=rotation[4]; m[9] =rotation[7]; m[13]=position[1]+p*(next_position[1]-position[1]);
   m[2]=rotation[2]; m[6]=rotation[5]; m[10]=rotation[8]; m[14]=position[2]+p*(next_position[2]-position[2]);
   m[3]=0.0f;        m[7]=0.0f;        m[11]=0.0f;        m[15]=1.0f;

   glPushMatrix();
    // uprava OpenGL MODELVIEW matice
    glMultMatrixf(m);
    // rekurzivne vykreslenie dalsich casti postavy
    DrawPlayer(link);
   glPopMatrix();
  }
 }
}
