#ifndef _MMD3MODEL_H_
#define _MMD3MODEL_H_

#include "mTexture.h"

#ifndef M_PI
#define M_PI 3.14159265359f
#endif

// hlavicka MD3 modelu
typedef struct
{
 char ident[4];      // ident=IDP3
 int version;        // verzia=MD3_VERSION (posledna ver. 15)
 char name[64];      // nazov MD3 modelu
 int flags;          // ?
 int numFrames;      // pocet snimkov (max. 1024)
 int numTags;        // pocet znaciek (max. 16)
 int numSurfaces;    // pocet sietovych modelov (max. 32)
 int numSkins;       // pocet povrchov
 int frames_start;   // zaciatok objektov Frames
 int tags_start;     // zaciatok objektov Tags
 int surfaces_start; // zaciatok objektov Surfaces
 int size;           // velkost hlavicky MD3 modelu
} md3_header_t;

// snimok
typedef struct
{
 float min[3];    // min. vrchol vymedzujucej kocky
 float max[3];    // max. vrchol vymedzujucej kocky
 float origin[3]; // pozicia vymedzujucej kocky
 float radius;    // polomer vymedzujucej gule
 char name[16];   // nazov snimku
} frame_t;

// znacka
typedef struct
{
 char name[64];    // nazov znacky
 float origin[3];  // suradnice znacky
 float axis[3][3]; // orientacia znacky
} tag_t;

// hlavicka sietoveho modelu
typedef struct
{
 char ident[4];       // ident=IDP3
 char name[64];       // nazov sietoveho modelu
 int flags;           // ?
 int numFrames;       // pocet snimkov
 int numShaders;      // pocet povrchov (max. 256)
 int numVerts;        // pocet vrcholov (max. 4096)
 int numTriangles;    // pocet trojuholnikov (max. 8192)
 int triangles_start; // zaciatok objektov Triangles
 int shaders_start;   // zaciatok objektov Shaders
 int st_start;        // zaciatok objektov ST (texture coordinates)
 int xyznormal_start; // zaciatok objektov Vertex
 int size;            // velkost hlavicky sietoveho modelu
} surface_header_t;

// povrchy
typedef struct
{
 char name[64];    // nazov shadera
 int shader_index; // cislo shadera
} shader_t;

// indexy na vrcholy trojuholnikov
typedef struct
{
 int indexes[3]; // indexy na 3 vrcholy
} triangle_t;

// texturovacie suradnice
typedef struct
{
 float st[2]; // ST (UV) texturovacie suradnice
} texcoord_t;

// vrcholy
typedef struct
{
 signed short int v[3]; // x,y,z suradnice vrcholov
 unsigned char n[2];    // kod normal vo vrcholoch
} vertex_t;

// sietovy model
typedef struct
{
 surface_header_t surface_h; // hlavicka sietoveho modelu
 shader_t *shaders;          // povrchy
 triangle_t *triangles;      // indexy na vrcholy trojuholnikov
 texcoord_t *texcoords;      // texturovacie suradnice
 vertex_t *vertices;         // vrcholy

 unsigned int tex_id;        // identifikator textury
 unsigned char is_texture;   // ma sietovy model texturu: 1 - ano 0 - nie
} surface_t;

// animacia
typedef struct
{
 // 4 ciselne udaje patriace animacii

 // 1. prvy snimok,2. pocet snimkov
 int first_frame,num_frames;
 // 3. pocet opakujucich sa snimkov,4. pocet snimkov za sekundu
 int looping_frames,frames_per_second;
} anim_t;

// typy pohybov
typedef enum {
 BOTH_DEATH1,   // 0
 BOTH_DEAD1,    // 1
 BOTH_DEATH2,   // 2
 BOTH_DEAD2,    // 3
 BOTH_DEATH3,   // 4
 BOTH_DEAD3,    // 5
 TORSO_GESTURE, // 6
 TORSO_ATTACK,  // 7
 TORSO_ATTACK2, // 8
 TORSO_DROP,    // 9
 TORSO_RAISE,   // 10
 TORSO_STAND,   // 11
 TORSO_STAND2,  // 12
 LEGS_WALKCR,   // 13
 LEGS_WALK,     // 14
 LEGS_RUN,      // 15
 LEGS_BACK,     // 16
 LEGS_SWIM,     // 17
 LEGS_JUMP,     // 18
 LEGS_LAND,     // 19
 LEGS_JUMPB,    // 20
 LEGS_LANDB,    // 21
 LEGS_IDLE,     // 22
 LEGS_IDLECR,   // 23
 LEGS_TURN,     // 24
 MAX_ANIMATIONS // 25
} animNumber_t;

// MD3 model
class MD3Model {
public:
 MD3Model();  // konstruktor
 ~MD3Model(); // destruktor

 // nacitanie obsahu MD3 suboru
 int LoadModel(char *filename);
 // nacitanie obrazkov pre povrchy sietovych modelov
 int LoadSkin(char *filename,char *imagepath);
 // nastavenie textury pre povrchy sietovych modelov
 int SetSurfaceTexture(char *surfacename,unsigned int TID);
 // nastavenie textury pre vsetky povrchy sietovych modelov
 //void SetAllSurfacesTexure(unsigned int TID);

 // prepojenia modelov
 int LinkModel(char *tagname,MD3Model *mod);
 // zrusenie prepojeni modelov
 //int UnLinkModel(char *tagname);

 // nastavenie informacii o snimku
 void SetFrameInfo(int f,int n,float p);
 // ziskanie informacii o snimku
 void GetFrameInfo(int *f,int *n,float *p);

 // vykreslenie 1 snimku
 void DrawModelOneFrame(int curFrame);
 // vykreslenie interpolovaneho snimku
 void DrawModelIFrame(int curFrame,int nextFrame,float p);
 // vykreslenie modelu (1 snimok alebo interpolovany snimok)
 void DrawModel(void);

 FILE *md3file;       // pre nacitanie md3 suboru
 md3_header_t header; // hlavicka MD3 suboru
 frame_t *frames;     // snimky
 tag_t *tags;         // znacky
 MD3Model **links;    // prepojenia
 surface_t *surfaces; // sietove modely

 int frame;     // aktualny snimok
 int nextframe; // nasledujuci snimok
 float ipol;    // interpolacny cas [0..1]

 // pomocne premenne
 int i,j;
 int cur_surface,cur_vertex,tri_num;
 int cur_offset_vertex,next_cur_offset_vertex;
 int au,av;
 float s,t,n[3],v[3],next_n[3],next_v[3];

 mTexture tt; // nacitanie textur pre povrchy sietovych modelov
};

// MD3 postava
class MD3Player {
public:
 MD3Player();  // konstruktor
 ~MD3Player(); // destruktor

 // inicializacia MD3 postavy
 void Init(void);
 // nastavenie parametrov animacie
 void SetAnim(int a);
 // vykreslenie MD3 postavy
 void Render(void);
private:
 // nacitanie MD3 postavy
 void LoadPlayer(char *dirname,char *skin,char *weapon);
 // nacitanie animacie
 int LoadAnim(char *filename);
 // vykreslenie vsetkych casti postavy - rekurzivna funkcia !!!
 void DrawPlayer(MD3Model *mod);

 // pomocne premenne
 int i,j;
 // casti MD3 postavy - MD3 modely (lower,upper,head,weapon)
 MD3Model l,u,h,w;
 // pre rekurzivne vykreslenie vsetkych casti postavy
 MD3Model *link;
 // 25 rozlicnych animacii
 anim_t anim[25];

 // parametre animacie
 int anim_l,anim_u;
 int l_startFrame,u_startFrame;
 int l_endFrame,u_endFrame;
 float l_fps,u_fps;

 // plynutie casu
 float aTime;
 int l_curFrame,u_curFrame;
 int l_nextFrame,u_nextFrame;
 float l_curTime,u_curTime;
 float l_oldTime,u_oldTime;

 // pre maticove vypocty
 float m[16];
 float *position,*rotation,*next_position,*next_rotation;
 // pre vypocet interpolacie
 float p;
};

#endif
